from ngsolve import *
from netgen.geom2d import SplineGeometry


#   point numbers 0, 1, ... 11
#   sub-domain numbers (1), (2), (3)
#  
#
#             7-------------6
#             |             |
#             |     (2)     |
#             |             |
#      3------4-------------5------2
#      |                           |
#      |             11            |
#      |           /   \           |
#      |         10 (3) 9          |
#      |           \   /     (1)   |
#      |             8             |
#      |                           |
#      0---------------------------1
#

def MakeGeometry():
    geometry = SplineGeometry()
    
    # point coordinates ...
    pnts = [ (0,0), (1,0), (1,0.6), (0,0.6), \
             (0.2,0.6), (0.8,0.6), (0.8,0.8), (0.2,0.8), \
             (0.5,0.15), (0.65,0.3), (0.5,0.45), (0.35,0.3) ]
    pnums = [geometry.AppendPoint(*p) for p in pnts]
    
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    lines = [ (0,1,1,1,0), (1,2,2,1,0), (2,5,2,1,0), (5,4,2,1,2), (4,3,2,1,0), (3,0,2,1,0), \
              (5,6,2,2,0), (6,7,2,2,0), (7,4,2,2,0), \
              (8,9,2,3,1), (9,10,2,3,1), (10,11,2,3,1), (11,8,2,3,1) ]
        
    for p1,p2,bc,left,right in lines:
        geometry.Append( ["line", pnums[p1], pnums[p2]], bc=bc, leftdomain=left, rightdomain=right)
    return geometry

mesh = Mesh(MakeGeometry().GenerateMesh (maxh=0.1))


fes = H1(mesh, order=2, dirichlet=[1], autoupdate=True)
u = fes.TrialFunction()
v = fes.TestFunction()

lam = CoefficientFunction([1, 1000, 10])
a = BilinearForm(fes, symmetric=False)
a += lam*grad(u)*grad(v)*dx

c = MultiGridPreconditioner(a, inverse = "sparsecholesky")


f = LinearForm(fes)
source = CoefficientFunction([0, 0, 1])
f += source*v*dx

gfu = GridFunction(fes, autoupdate=True)
Draw (gfu)

gferr = GridFunction(L2(mesh, order=0, autoupdate=True), autoupdate=True)
Draw (gferr, mesh, "err")

def SolveBVP():
    a.Assemble()
    f.Assemble()
    inv = CGSolver(a.mat, c.mat)
    gfu.vec.data = inv * f.vec
    Redraw (blocking=True)





def CalcError():
    h = specialcf.mesh_size
    n = specialcf.normal(mesh.dim)
    def Laplace(u): return Trace(u.Operator("hesse"))
    elerr = Integrate ( (h*(lam*Laplace(gfu)+source))**2*dx + h*(n*(lam*grad(gfu)-(lam*grad(gfu)).Other()))**2*dx(element_boundary=True), mesh, element_wise=True)

    gferr.vec.data.FV()[:] = elerr
    Redraw (blocking=True)    

    maxerr = max(elerr)
    print ("maxerr = ", maxerr)
    for el in mesh.Elements():
        mesh.SetRefinementFlag(el, elerr[el.nr] > 0.25*maxerr)


with TaskManager():
    while fes.ndof < 20000:  
        SolveBVP()
        CalcError()
        mesh.Refine()
    
SolveBVP()
CalcError()
    
